/*
 * Convert tux binary log files into W3C-conform HTTPD log files
 *
 * tux log files are ~50% more space and IO bandwith effective
 * than the ASCII version.
 *
 * Copyright 2000 Red Hat, Inc.
 * Written by Ingo Molnar <mingo@redhat.com>
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/wait.h>
#include <linux/unistd.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/types.h>

#define MAX_URL_LEN 4096

/* */
#ifdef __i386__
#define O_LARGEFILE 0100000
#endif

#ifdef __alpha__
#define O_LARGEFILE 0400000
#endif

#ifdef __ia64__
#define O_LARGEFILE    0100000
#endif

#ifdef __powerpc__
#define O_LARGEFILE 0200000
#endif

#ifdef __s390__
#define O_LARGEFILE 0100000
#endif

#ifdef _BIG_ENDIAN
#define LOG_ID1 0x1111beef
#define LOG_ID2 0x2222beef
#define LOG_ID3 0x1112beef
#define LOG_ID4 0x2223beef
#else
#define LOG_ID1 0xefbe1111
#define LOG_ID2 0xefbe2222
#define LOG_ID3 0xefbe1211
#define LOG_ID4 0xefbe2322
#endif

typedef struct standard_tux_record {
	__u32 ip_addr;
	__u32 time;
	__u32 bytes;
	__u32 http_status;
} tux_std_rec_t;

typedef struct extended_tux_record {
	__u32 ip_addr;
	__u32 ip_port;
	__u32 time;
	__u32 time_accept;
	__u32 time_parse;
	__u32 time_output;
	__u32 time_flush;
	__u32 had_cachemiss;
	__u32 keep_alive;
	__u32 bytes;
	__u32 http_status;
} tux_ext_rec_t;

#include "myio.c"

static int parse_string (int file, char *str)
{
	int c;
	unsigned int i = 0;

	for (;;) {
		c = my_getc(file);
		if (c == EOF)
			return -1;
		str[i] = c;
		if (++i >= MAX_URL_LEN)
			break;
		if (!c)
			break;
	}
	return 0;
}

/*
 * The server logs the following information for each request made:
 * address of the requestor, a date and time stamp accurate to at
 * least 1 second, specification of the file requested, size of the
 * file transferred, and the final status of the request.
 */

int main (int argc, char **argv)
{
	int i = 0, count = 0, c, len, extended, countonly, nr_records, combined;
	char method_str [MAX_URL_LEN];
	char uri_str [MAX_URL_LEN];
	char query_str [MAX_URL_LEN];
	char user_agent_str [MAX_URL_LEN];
	char referer_str [MAX_URL_LEN];
	__u32 id = 0;
	char timestr [100];
	char *filename;
	tux_std_rec_t std_rec;
	tux_ext_rec_t ext_rec;
	int file;

	if ((argc != 2) && (argc != 3)) {
out_usage:
		fprintf(stderr, "usage: tux2w3c [-x|-c|-b] <tux-logfile>\n");
		fprintf(stderr, " -x eXtended log format\n");
		fprintf(stderr, " -c Count only\n");
		fprintf(stderr, " -b brief\n");
			
		exit(-1);
	}
	extended = 0;
	combined = 1;
	countonly = 0;
	filename = argv[1];
	if (argc == 3) {
		if (!strcmp(argv[1], "-x")) {
			extended = 1;
			combined = 0;
			filename = argv[2];
		} else
		if (!strcmp(argv[1], "-c")) {
			countonly = 1;
			combined = 0;
			filename = argv[2];
		} else
		if (!strcmp(argv[1], "-b")) {
			combined = 0;
			filename = argv[2];
		} else
			goto out_usage;
	}

	file = open(filename, O_RDONLY|O_LARGEFILE);
	if (!file) {
		fprintf(stderr, "could not open TUX log file '%s'!\n", filename);
		exit(-1);
	}
	nr_records = 0;
	for (;;) {
		int extended_rec, referer_log;
		__u32 ip_addr;
		unsigned long time;
		struct tm *st_localtime;
		__u32 bytes;
		__u32 http_status;

		c = my_getc(file);
		count++;
		if (c == EOF)
			break;
		id = (id << 8) | c;
		if ((id != LOG_ID1) && (id != LOG_ID2) &&
				(id != LOG_ID3) && (id != LOG_ID4))
			continue;
		extended_rec = 0;
		if ((id == LOG_ID2) || (id == LOG_ID4))
			extended_rec = 1;
		referer_log = 0;
		if ((id == LOG_ID4) || (id == LOG_ID3))
			referer_log = 1;
		if (!extended_rec) {
			if (my_read(file, &std_rec, sizeof(std_rec)) != sizeof(std_rec))
				break;
			ip_addr = std_rec.ip_addr;
			time = std_rec.time;
			bytes = std_rec.bytes;
			http_status = std_rec.http_status;
		} else {
			if (my_read(file, &ext_rec, sizeof(ext_rec)) != sizeof(ext_rec))
				break;
			ip_addr = ext_rec.ip_addr;
			time = ext_rec.time;
			bytes = ext_rec.bytes;
			http_status = ext_rec.http_status;
		}
		i = 0;
		if (parse_string(file, method_str))
			break;
		if (parse_string(file, uri_str))
			break;
		if (parse_string(file, query_str))
			break;
		if (extended_rec)
			if (parse_string(file, user_agent_str))
				break;
		referer_str[0] = 0;
		if (referer_log)
			if (parse_string(file, referer_str))
				break;

		nr_records++;
		if (countonly)
			continue;

		printf("%u.%u.%u.%u",
			*(((unsigned char *)&ip_addr) + 0),
			*(((unsigned char *)&ip_addr) + 1),
			*(((unsigned char *)&ip_addr) + 2),
			*(((unsigned char *)&ip_addr) + 3));
		
		if (extended && extended_rec) 
			printf(":%d", ext_rec.ip_port);
		
		printf(" - -");

		if (combined) {
			st_localtime = localtime(&time);
			strftime(timestr, 99, "%d/%b/%Y:%H:%M:%S %z", st_localtime);
		} else
			strcpy(timestr, ctime(&time));
			
		len = strlen(timestr);
		if (!combined)
			timestr[len-1] = 0;
		
		if (combined)
			printf(" [%s]", timestr);
		else
			printf(" %s", timestr);
		
		if (extended && extended_rec)
			printf(" (%d,%d,%d)",
				ext_rec.time_parse-ext_rec.time_accept,
			 	ext_rec.time_output ?  ext_rec.time_output -
					ext_rec.time_accept : -1,
				ext_rec.time_flush-ext_rec.time_accept);

		if (extended && extended_rec)
			printf(" [K:%d,CM:%d]",
				ext_rec.keep_alive, ext_rec.had_cachemiss);
		
		printf(" \"%s", method_str);
		printf(" %s", uri_str);
		printf(" %s\"", query_str);

		if (combined)
			printf(" %d %d", http_status, bytes);
		else
			if (extended && extended_rec)
				printf(" \"%s\"", user_agent_str);

		if (referer_str[0])
			printf(" \"%s\"", referer_str);
		else
			if (combined)
				printf(" \"-\"");
			else
				printf(" -");

		if (combined)
			printf(" \"%s\"", user_agent_str);
		else
			printf(" %d %d", bytes, http_status);
		
		printf("\n");
	}
	if (countonly)
		printf("nr_records: %d\n", nr_records);
	return 0;
}

